/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

@namespace url("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul");

button {
  -moz-appearance: button;
  /* The horizontal margin used here come from the Aqua Human Interface
     Guidelines, there should be 12 pixels between two buttons. */
  margin: 5px 6px 3px;
  min-width: 79px;
  color: ButtonText;
  text-shadow: none;
}

.button-text {
  margin: 1px 0 !important;
  -moz-margin-start: 3px !important;
  -moz-margin-end: 2px !important;
  text-align: center;
}

.button-icon {
  -moz-margin-start: 1px;
}

button[type="default"] {
  font: menu;
}

/* .......... disabled state .......... */

button[disabled="true"] {
  color: GrayText;
}

/* ::::: menu/menu-button buttons ::::: */

button[type="menu-button"] {
  margin: 0;
  border: none;
}

.button-menu-dropmarker,
.button-menubutton-dropmarker {
  -moz-appearance: none !important;
  border: none;
  background-color: transparent !important;
  margin: 1px;
}

.button-menu-dropmarker {
  display: none;
}

/* ::::: plain buttons ::::: */

button.plain {
  margin: 0 !important;
  padding: 0 !important;
}

/* ::::: help button ::::: */

button[dlgtype="help"] {
  min-width: 1px;
  padding: 0;
  -moz-appearance: none;
  -moz-box-align: start;
  height: 24px;
  width: 24px;
  margin: 3px 4px;
}

button[dlgtype="help"][disabled] {
  opacity: 0.5;
}

button[dlgtype="help"]:-moz-focusring {
  outline: 2px solid -moz-mac-focusring;
  outline-offset: -2px;
  -moz-outline-radius: 10000px;
}

button[dlgtype="help"] > .button-box > .button-icon {
  list-style-image: url("chrome://global/skin/icons/question-mark.png");
  -moz-image-region: rect(0 24px 24px 0);
  padding: 0;
  margin: 0;
}

button[dlgtype="help"]:active > .button-box > .button-icon {
  -moz-image-region: rect(0 48px 24px 24px);
}

button[dlgtype="help"] > .button-box > .button-text {
  display: none;
}
